use mastodon_async::entities::visibility::Visibility;
use mastodon_async::helpers::{cli, toml as masto_toml};
use mastodon_async::prelude::*;
use serde::{Deserialize, Serialize};
use std::io::Cursor;
use std::process::exit;
use std::time::Duration;

#[derive(Deserialize)]
struct ImagesWrap {
  images: Vec<Image>,
}

#[derive(Deserialize, Clone, Eq, PartialEq, Debug)]
struct Image {
  url: String,
  _id: String,
  status: String,
}

#[derive(Serialize)]
struct ImageStatus {
  status: String,
}

#[derive(Serialize)]
struct AccountUpdate {
  note: String,
}

const CONFIG_FILENAME: &str = "config.toml";

type DynResult<T> = Result<T, Box<dyn std::error::Error>>;

#[derive(Deserialize, Serialize, Default)]
struct Config {
  bot: Bot,
  backend: Backend,
  errors: Errors,
  files: Files,
}

#[derive(Deserialize, Serialize, Default)]
struct Files {
  tempfile: String,
}

#[derive(Deserialize, Serialize, Default)]
struct Bot {
  name: String,
  instance: String,
  bio: String,
}

#[derive(Deserialize, Serialize, Default)]
struct Backend {
  url: String,
  app: String,
  secret: String,
}

#[derive(Deserialize, Serialize, Default)]
struct Errors {
  maintainers: String,
  out_of_images: String,
  retry: u8,
}

#[tokio::main] // requires `features = ["mt"]
async fn main() -> DynResult<()> {
  stderrlog::new()
    .module(module_path!())
    .quiet(false)
    .verbosity(4) // Debug
    .timestamp(stderrlog::Timestamp::Second)
    .init()?;

  let config: Config = get_config();
  let mastodon = get_account(&config).await;

  match get_next_url(&config).await {
    Ok(image) => match image {
      Some(image) => {
        let mut retry: u8 = 0;
        while let Err(err) = post_image(&mastodon, &image.url, &config, Visibility::Unlisted).await
        {
          log::warn!("Cannot post image, retry: {}, {}", retry, err);
          async_std::task::sleep(Duration::new(1, 0)).await;
          retry += 1;
          if retry >= config.errors.retry {
            log::error!("Max ammount of retries reached on post_image");
            log::info!("Reverting file changes");
            exit(1);
          }
        }
        set_url_as_posted(&config, &image).await?;
        let mut retry: u8 = 0;
        while let Err(err) = update_bio(&mastodon, &config).await {
          log::warn!("Cannot update bio, retry: {}, {}", retry, err);
          async_std::task::sleep(Duration::new(1, 0)).await;
          retry += 1;
          if retry >= config.errors.retry {
            log::error!("Max ammount of retries reached on update bio");
            exit(1);
          }
        }
      }
      None => {
        let mut retry: u8 = 0;
        while let Err(err) = post(
          &mastodon,
          &format!(
            "{} {}",
            &config.errors.maintainers, &config.errors.out_of_images
          ),
          Visibility::Direct,
        )
        .await
        {
          log::warn!("Cannot post, retry: {}, {}", retry, err);
          async_std::task::sleep(Duration::new(1, 0)).await;
          retry += 1;
          if retry >= config.errors.retry {
            log::error!("Max ammount of retries reached on post");
            exit(1);
          }
        }
      }
    },
    Err(err) => {
      log::error!("Cannot get next image: {}", err);
      match post(
        &mastodon,
        &format!("{} {}", &config.errors.maintainers, &err.to_string()),
        Visibility::Direct,
      )
      .await
      {
        Ok(_) => {}
        Err(err) => {
          log::error!("Cannot post error message: {}", err);
          exit(1);
        }
      };
    }
  }
  Ok(())
}

fn get_config() -> Config {
  match parse_config(CONFIG_FILENAME) {
    Ok(config) => config,
    Err(err) => {
      log::error!("Config file parsing unsuccesful: {}", err);
      generate_config().unwrap();
      log::info!("Please check the working directory to find a new config file");
      exit(1);
    }
  }
}

fn generate_config() -> DynResult<()> {
  let config = Config::default();
  let config = toml::to_string(&config)?;
  std::fs::write(CONFIG_FILENAME, config)?;
  Ok(())
}

/// Parses the given filename to a config struct
fn parse_config(filename: &str) -> DynResult<Config> {
  let toml_file = std::fs::read_to_string(filename)?; //.expect("No config file, consider getting the original one and modifing it");
  Ok(toml::from_str(&toml_file)?) //("Malformed config file, check the original one for reference")
}

async fn get_account(config: &Config) -> Mastodon {
  if let Ok(data) = masto_toml::from_file("mastodon-data.toml") {
    Mastodon::from(data)
  } else {
    match register(config).await {
      Ok(account) => account,
      Err(err) => {
        log::error!("Api registation unsuccesful: {}", err);
        exit(1);
      }
    }
  }
}

async fn get_next_url(config: &Config) -> DynResult<Option<Image>> {
  let images: ImagesWrap = reqwest::get(format!(
    "{}/images?status=available&limit=1",
    config.backend.url
  ))
  .await?
  .json()
  .await?;

  let images = images.images;

  if images.is_empty() {
    Ok(None)
  } else {
    Ok(Some(images[0].clone()))
  }
}

#[derive(Serialize)]
struct Auth {
  app: String,
  secret: String,
}

#[derive(Deserialize)]
struct Token {
  token: String,
}

async fn set_url_as_posted(config: &Config, image: &Image) -> DynResult<()> {
  let client = reqwest::Client::new();

  let auth = &Auth {
    app: config.backend.app.to_string(),
    secret: config.backend.secret.to_string(),
  };

  let token: Token = client
    .post(&format!("{}/login", config.backend.url))
    .json(auth)
    .send()
    .await?
    .json()
    .await?;

  let response = client
    .put(&format!("{}/images/{}", config.backend.url, image._id))
    .bearer_auth(token.token)
    .json(&ImageStatus {
      status: "consumed".to_string(),
    })
    .send()
    .await?;
  if response.status().is_success() {
    log::info!("Set url {} as posted", image.url);
    Ok(())
  } else {
    let error = response.text().await?;
    log::error!("Set url response: {}", error);
    Err(error.into())
  }
}

async fn post_image(
  account: &Mastodon,
  url: &String,
  config: &Config,
  visibility: Visibility,
) -> DynResult<Status> {
  fetch_url(url, &config.files.tempfile).await?; //.expect("Error fetching url");
  let attachment = account
    .media(&config.files.tempfile, Some(url.to_string()))
    .await?; //.expect("Attachment upload error");
  let attachment = account
    .wait_for_processing(attachment, Default::default())
    .await?; //.expect("Attachment processing error");
  let status = StatusBuilder::new()
    .media_ids(&[attachment.id])
    .visibility(visibility)
    .sensitive(true)
    .build()?; //.expect("Could not build status"); // we should retry
  let status = account.new_status(status).await?; //.expect("Error generating status"); // we should retry or delete last url in posted
  log::info!("Image status posted: {}", url);
  Ok(status)
}

async fn update_bio(account: &Mastodon, config: &Config) -> DynResult<()> {
  let images: ImagesWrap = reqwest::get(format!("{}/images?status=available", config.backend.url))
    .await?
    .json()
    .await?;

  let remaining = images.images.len();
  let client = reqwest::Client::new();

  let account_update = AccountUpdate {
    note: format!("{}\n\n{} new images remaining", config.bot.bio, remaining),
  };

  let response = client
    .patch(&format!(
      "{}/api/v1/accounts/update_credentials",
      config.bot.instance
    ))
    .bearer_auth(&account.data.token)
    .json(&account_update)
    .send()
    .await?;
  log::info!("Bio updated, response status: {}", response.status());
  Ok(())
}

async fn post(account: &Mastodon, msg: &str, visibility: Visibility) -> DynResult<Status> {
  let status = StatusBuilder::new()
    .visibility(visibility)
    .status(msg)
    .build()?; //.expect("Error building error status");
  let post = account.new_status(status).await?; //.expect("Error posting error status");
  log::info!("Text status posted: {}", msg);
  Ok(post)
}

async fn fetch_url(url: &String, file_name: &String) -> DynResult<()> {
  let response = reqwest::get(url);
  let mut file = std::fs::File::create(file_name)?;
  let mut content = Cursor::new(response.await?.bytes().await?);
  std::io::copy(&mut content, &mut file)?;
  Ok(())
}

async fn register(config: &Config) -> DynResult<Mastodon> {
  let registration = Registration::new(&config.bot.instance)
    .client_name(&config.bot.name)
    .scopes(Scopes::all())
    .build()
    .await?;
  let mastodon = cli::authenticate(registration).await?;

  // Save app data for using on the next run.
  masto_toml::to_file(&mastodon.data, "mastodon-data.toml")?;

  Ok(mastodon)
}

#[cfg(test)]
mod tests {

  use reqwest::StatusCode;

  use super::*;
  const TMPTESTDIR: &str = "/tmp/botimage";
  const TEST_URL: &str = "https://2.gravatar.com/avatar/be8eb8426d68e4beb50790647eda6f6b";

  #[tokio::test]
  async fn fetch_url_should_fetch() {
    fetch_url(&TEST_URL.to_string(), &TMPTESTDIR.to_string())
      .await
      .unwrap();
    std::fs::read(TMPTESTDIR).unwrap();
  }

  #[tokio::test]
  async fn post_should_post() {
    let client = reqwest::Client::new();
    let config = get_config();
    let account = get_account(&config).await;
    let msg = "Test!".to_string();

    let status = post(&account, &msg, Visibility::Direct).await.unwrap();
    let response = client
      .get(dbg!(format!(
        "{}/api/v1/statuses/{}",
        &config.bot.instance,
        &status.id.to_string()
      )))
      .bearer_auth(dbg!(&account.data.token))
      .send()
      .await
      .unwrap();
    account.delete_status(&status.id).await.unwrap();

    assert_eq!(response.status(), StatusCode::OK)
  }

  #[tokio::test]
  async fn post_image_works() {
    let client = reqwest::Client::new();
    let config = get_config();
    let account = get_account(&config).await;

    let status = post_image(&account, &TEST_URL.to_string(), &config, Visibility::Direct)
      .await
      .unwrap();

    let response = account.get_status(&status.id).await.unwrap();

    account.delete_status(&status.id).await.unwrap();
    let attachment = &response.media_attachments[0];

    let response = client
      .get(attachment.url.clone().unwrap())
      .bearer_auth(dbg!(&account.data.token))
      .send()
      .await
      .unwrap();

    assert_eq!(response.status(), StatusCode::OK)
  }

  #[tokio::test]
  async fn set_as_posted_works() {
    let config = get_config();
    const IMAGE: &str = "https://picsum.photos/id/2";
    let expected = insert_image(&config, IMAGE).await.unwrap();

    set_url_as_posted(&config, &expected).await.unwrap();
    let image: ImageWrap = reqwest::get(format!("{}/images/{}", config.backend.url, expected._id))
      .await
      .unwrap()
      .json()
      .await
      .unwrap();
    let image = image.image;
    assert_eq!(image.status, "consumed");
  }

  #[tokio::test]
  async fn get_next_url_works() {
    let config = get_config();

    const IMAGE: &str = "https://picsum.photos/id/1";
    let expected = insert_image(&config, IMAGE).await.unwrap();

    // Get test url
    let image = get_next_url(&config).await.unwrap().unwrap();
    assert_eq!(image.url, IMAGE);

    set_url_as_posted(&config, &expected).await.unwrap();
    let image: ImageWrap = reqwest::get(format!("{}/images/{}", config.backend.url, expected._id))
      .await
      .unwrap()
      .json()
      .await
      .unwrap();
    let image = image.image;
    assert_eq!(image.status, "consumed");

    // Test that now it does not get it
    let image = get_next_url(&config).await.unwrap();
    assert_eq!(image, None);
  }

  async fn insert_image(config: &Config, url: &str) -> DynResult<Image> {
    let client = reqwest::Client::new();

    let auth = &Auth {
      app: config.backend.app.to_string(),
      secret: config.backend.secret.to_string(),
    };

    let token: Token = client
      .post(&format!("{}/login", config.backend.url))
      .json(auth)
      .send()
      .await?
      .json()
      .await?;

    let response = client
      .post(&format!("{}/images", config.backend.url))
      .bearer_auth(&token.token)
      .json(&Insert {
        url: url.to_string(),
        status: "available".to_string(),
        tags: vec![],
      })
      .send()
      .await?;

    if response.status().is_success() {
      let response: ImageWrap = response.json().await?;
      return Ok(response.image);
    }
    dbg!(response.text().await?);
    panic!();
  }

  #[derive(Serialize)]
  struct Insert {
    url: String,
    status: String,
    tags: Vec<String>,
  }

  #[derive(Deserialize)]
  struct ImageWrap {
    image: Image,
  }
}
